

from contacts24.errors import input_error



@input_error
def add_note(args, note_book):
    """Adds a note to the notebook."""
    note_id, title, text = args
    note_book.add_note(note_id, title, text)
    return "Note added."


@input_error
def change_note(args, note_book):
    """Changes the text of an existing note."""
    key, new_text = args
    note_book.change_note(int(key), new_text)
    return "Note changed."


@input_error
def get_notes(note_book):
    """Shows all notes from the notebook."""
    notes = note_book.get_notes()
    return "\n".join([str(note) for note in notes])


@input_error
def delete_note(args, note_book):
    """Deletes a note from the notebook."""
    key, = args
    note_book.delete_note(int(key))
    return "Note deleted."


@input_error
def search_text(args, note_book):
    """Searches for notes by a text query."""
    search_query, = args
    found_notes = note_book.search_text(search_query)
    if found_notes:
        return "\n".join([f"Note ID: {note_id}, Note Data: {note_data}" for note_id, note_data in found_notes])
    else:
        return "No notes found."


from collections import defaultdict
from datetime import date, datetime, timedelta
from typing import DefaultDict, List

from contacts24.config import DATE_FORMAT
from contacts24.models.address_book import AddressBook


def get_birthdays_within_days(address_book: AddressBook, within_days: int) -> str:
    """
    Generates a greeting string for the people whose birthdays are within the given number of days.

    Args:
        address_book (AddressBook): The address book containing the contacts.
        within_days (int): The number of upcoming days to check for birthdays.

    Returns:
        str: The greeting string containing birthday wishes for the people whose birthdays are coming up within the surveyed days.

    """
    relative_date = _get_relative_date(within_days)
    birthday_map = _construct_birthday_map(address_book, relative_date, within_days)

    return _construct_greeting_string(relative_date, birthday_map, within_days)


def _construct_birthday_map(address_book: AddressBook, relative_date: date, within_days: int) -> DefaultDict[str, List[str]]:
    birthday_map = defaultdict()

    # Initialize an ordered dictionary with date keys for within_days
    ordered_keys = []
    for i in range(within_days):
        current_date_key = f"In {i} days ({(relative_date + timedelta(days=i)).strftime(DATE_FORMAT)})"
        ordered_keys.append(current_date_key)
        birthday_map[current_date_key] = []

    for colleague in address_book.data.values():
        name = colleague.name.value

        if not colleague.birthday:
            continue

        birthday = colleague.birthday.get_birthday_datetime()
        birthday_this_year = _get_birthday_this_year(birthday, relative_date)
        
        delta_days = (birthday_this_year - relative_date).days

        if 0 <= delta_days < within_days:
            current_date_key = f"In {delta_days} days ({(relative_date + timedelta(delta_days)).strftime(DATE_FORMAT)})"
            birthday_map[current_date_key].append(
                f"{name}"
            )

    birthday_map = {k: v for k, v in birthday_map.items() if v}

    return birthday_map

def _get_birthday_this_year(birthday: date, relative_date: date) -> date:
    birthday_this_year = birthday.replace(year=relative_date.year)
    
    if birthday_this_year < relative_date:
        birthday_this_year = birthday_this_year.replace(year=relative_date.year + 1)

    return birthday_this_year


def _get_relative_date(within_days: int) -> date:
    return (datetime.now()).date()


def _construct_greeting_string(relative_date: date, birthday_map: DefaultDict[str, List[str]], within_days: int) -> str:
    greeting_string = f"---\nColleagues to greet for the next {within_days} days, as of {relative_date.strftime(DATE_FORMAT)}:\n---"
    
    if birthday_map:
        for week_day in birthday_map:
            week_day_birthdays = birthday_map[week_day]
            if week_day_birthdays:
                greeting_string += f"\n{week_day}: {', '.join(week_day_birthdays)}"
    else:
        greeting_string += "\nNo birthdays this week :("

    return greeting_string







from collections import UserDict
from contacts24.errors import NonExistingNote



class Notes(UserDict):
    """This class provides functionality for working with notes"""
    def add_note(self, title, text):
        """Adds a new note to the notebook."""
        max_index = max(self.data.keys()) if self.data else 0
        note_id = max_index + 1  # Generates a unique identifier for the note
        self.data[note_id] = {'note_id': note_id, 'title': title, 'text': text}

    def change_note(self, key, new_text):
        """Changes the text of an existing note by its key."""
        if key in self.data:
            self.data[key]['text'] = new_text
        else:
            raise NonExistingNote()

    def get_notes(self):
        """Returns all notes from the notebook."""
        return self.data.values()

    def delete_note(self, key):
        """Deletes a note by its key."""
        if key in self.data:
            del self.data[key]
        else:
            raise NonExistingNote()

    def search_text(self, search_query):
        """Searches for notes by a text query."""
        found_notes = []
        for note_id, note_data in self.data.items():
            if search_query in note_data['text']:
                found_notes.append((note_id, note_data))
        return found_notes 
    


from collections import UserDict
from typing import Optional

from contacts24.models.record import Record
from contacts24.config import ADDRESSBOOK_FILE


class AddressBook(UserDict):
    def add_record(self, record: Record) -> None:
        self.data[record.name.value] = record

    def find(self, name: str) -> Optional[Record]:
        return self.data.get(name)

    def delete(self, name: str) -> None:
        if name in self.data:
            del self.data[name]
            print(f"{name} record is deleted")
        else:
            print(f"No records found by the name {name}")

    def save_to_file(self, filepath: str = ADDRESSBOOK_FILE) -> None:
        from contacts24.db import save_address_book
        if filepath:
            save_address_book(self, filepath)
        else:
            save_address_book(self)

    @staticmethod
    def load_from_file(filepath: str = ADDRESSBOOK_FILE) -> "AddressBook":
        from contacts24.db import get_contacts
        if filepath:
            return get_contacts(filepath)
        else:
            return get_contacts()

    def __str__(self) -> str:
        # Initialize the header and its length
        columns = ['Name', 'Birthday', 'Address', 'Phones', 'Emails']
        col_width = [len(h)+ 3 for h in columns]

        # Gather all data
        data = []
        for name, record in self.data.items():
            phones = ', '.join(str(p.value) for p in record.phones) if record.phones else 'N/A'
            emails = ', '.join(str(e.value) for e in record.emails) if record.emails else 'N/A'
            birthday = str(record.birthday) if record.birthday else 'N/A'
            address = str(record.address) if record.address else 'N/A'
            row = [name, birthday, address, phones, emails]
            
            # Update maximum column width
            col_width = [max(col_width[i], len(cell)) for i, cell in enumerate(row)]

            data.append(row)

        # Define format string for each row
        format_string = ''.join(f'{{:<{w}}} ' for w in col_width)

        # Build table string
        output = format_string.format(*columns) + '\n'
        output += '\n'.join(format_string.format(*row) for row in data)

        return output



import os

def get_file_path(filename = "") -> str:
    """Get user home dir

    Args:
        filename (str): file name to get home path

    Returns:
        str: Path to user dir, else, current working dir
    """
    try:
        folder = os.path.expanduser('~')
    except Exception:
        folder = os.getcwd()
    
    if filename and filename != '':
        folder = os.path.join(folder , filename)

    return folder


import json

from contacts24.models.address_book import AddressBook
from contacts24.models.notes import Notes
from contacts24.config import ADDRESSBOOK_FILE, NOTES_FILE
from contacts24.errors import app_error_wrapper
from contacts24.serialization_helper import record_deserialization, record_serialization, note_deserialization, note_serialization


#region AddressBook

@app_error_wrapper
def get_contacts(filename: str = ADDRESSBOOK_FILE) -> AddressBook:
    """Get addressbook from json file (default config.ADDRESSBOOK_FILE)

    Returns:
        AddressBook: AddressBook from json
    """
    address_book = AddressBook()
    
    with open(filename, "r") as file:
        records_list = json.load(file)
        
    for record_dict in records_list:
        record = record_deserialization(record_dict)
        address_book.add_record(record)
    
    return address_book

@app_error_wrapper
def save_address_book(address_book: AddressBook, filename: str = ADDRESSBOOK_FILE) -> None:
    """Save address book to json file (from config.ADDRESSBOOK_FILE)

    Args:
        address_book (AddressBook): Address Book to save
        filename (str): file name

    """
    with open(filename, "w") as file:
        records_list = [record_serialization(record) for record in address_book.data.values()]
        json.dump(records_list, file)

#endregion

#region Notes

@app_error_wrapper
def get_notes(filename: str = NOTES_FILE) -> Notes:
    """Get notes from fixed json file (from config.NOTES_FILE)

    Returns:
        Notes: List of notes
    """
    notes = Notes()
    
    with open(filename, "r") as file:
        notes_list = json.load(file)
        for note_dict in notes_list:
            note = note_deserialization(note_dict)
            notes.add_note(note)
    
    return notes

@app_error_wrapper
def save_notes(notes: Notes, filename: str = NOTES_FILE) -> None:
    """Save notes to json file (from config.NOTES_FILE)

    Args:
        notes (Notes): Notes to save
    """

    with open(filename, "w") as file:
        notes_list = [note_serialization(note) for note in notes.data.values()]
        json.dump(notes_list, file)

#endregion